/* ccgo: igs.cc
 *
 * Igs class definition
 * 
 * Copyright 2002 Chun-Chung Chen <cjj@u.washington.edu>
 *
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <iostream>
#include "igs.hh"
#include "igsmatch.hh"
#include <sstream>
#define atms(i, j) (rx.match[i].atom[j].str())
#define atmi(i, j) atoi(rx.match[i].atom[j].str().c_str())
#define atmf(i, j) atof(rx.match[i].atom[j].str().c_str())

using namespace std;

Igs::Igs() :
	Gtk::Window(GTK_WINDOW_TOPLEVEL),
	catch_enter(NULL)
{
	Gtk::VBox * v;
	Gtk::HBox * h;
	Gtk::Button * b;
	Gtk::HPaned * p;
	parser = & Igs::ps_init;
	cons.set_input_active(false);
	stage = 0;
	game_board = NULL;
	h = manage(new Gtk::HBox);
	b = manage(new Gtk::Button("quit"));
	b->clicked.connect(destroy.slot());
	h->pack_start(* b, false, false);
	b->show();
	v = manage(new Gtk::VBox);
	v->pack_start(data);
	data.show();
	v->pack_start(* h, false, false);
	h->show();
        p = manage(new Gtk::HPaned);
	p->add1(cons);
	cons.show();
	p->add2(* v);
	v->show();
	add(* p);
	p->show();
	cons.enter.connect(slot(this, & Igs::enter_command));
	conn.conn_lost.connect(destroy.slot());
	conn.line_ready.connect(slot(this, & Igs::grab_line));
	data.observe_close.connect(slot(this, & Igs::unob_game));
	data.move_skips.connect(slot(this, & Igs::moves_query));
	data.game_query.connect(slot(this, & Igs::game_query));
	data.choose_game.connect(slot(this, & Igs::observe_game));
	data.choose_player.connect(slot(this, & Igs::stats_player));
	data.match_put.connect(slot(this, & Igs::send_move));
	data.match_undo.connect(slot(this, & Igs::send_undo));
	data.match_pass.connect(slot(this, & Igs::send_pass));
	data.match_done.connect(slot(this, & Igs::send_done));
	data.match_talk.connect(slot(this, & Igs::send_say));
	data.game_talk.connect(slot(this, & Igs::send_kibitz));
	data.player_list_refresh.connect(slot(this, & Igs::player_list_query));
	data.game_list_refresh.connect(slot(this, & Igs::game_list_query));
}

void Igs::ps_init(const string & s)
{
	
	if (rx.exec(s, "^Login:") || rx.exec(s, "^1 0")) {
		parser = & Igs::ps_login;
		stage = 0;
		ps_login(s);
	} else cons.dump(s, IgsConsole::TT_FIXED);
}

void Igs::ps_login(const string & s)
{
	if (rx.exec(s, "^Login:") || rx.exec(s, "^1 0")) {
		stage = 1;
		cons.set_prompt("Login:");
		cons.set_visible(true);
		cons.set_input_active(true);
		catch_enter = & username;
	} else if (rx.exec(s, "^Password:") || rx.exec(s, "^1 1")) {
		stage = 2;
		cons.set_prompt("Password:");
		cons.set_visible(false);
	} else if (rx.exec(s, "#>")) {
		if (stage < 4) {
			stage = 4;
			// cerr << "get normal prompt" << endl;
			conn.send_line("toggle client 1");
		} else {
			//  can't enter client mode!!!
			cerr << "Can't enter client mode!!!" << endl;
			throw string("Client Mode Error");
		}
	} else if (rx.exec(s, "^1 [2-9]")) {
		if (stage < 5) {
			conn.send_line("toggle quiet 0");
			cons.clear_history();
			cons.set_visible(true);
			stage = 5;
		} else {
			parser = & Igs::ps_prompt;
			stage = 0;
			data.set_myid(username);
			ostringstream t;
			t << "IGS(" << username << ")";
			set_title(t.str());
			// cerr << "sending 'games' command" << endl;
			queue_command("games");
			queue_command("who");
		}
	} else if (rx.exec(s, "^9 File")) { // motd
		if (stage == 3) stage = 2;
		else stage = 3;
	} else if (rx.exec(s, "^(\\d+) \\{")) {
		// unimportant information at this time...
	} else if (stage == 3) {
		cons.dump(s, IgsConsole::TT_FILE);
	}
}

void Igs::ps_prompt(const string & s)
{
	int cmd;
	string dat;
	if (rx.exec(s, "^(\\d+)\\s+(.*)") == 0) {
		if (s != "") cons.dump(s);
		return;
	}
	cmd = atmi(0, 0);
	dat = atms(0, 1);
	// cerr << "(" << cmd << "):" << dat << endl;
	switch (cmd) {
	case 1: // prompt
		switch (atoi(dat.c_str())) {
		case 5: // normal
			cons.set_prompt("#>");
			break;
		case 6: // in a match
			cons.set_prompt("M>");
			break;
		case 7: // scoring
			cons.set_prompt("S>");
			break;
		case 8: // observing games
			cons.set_prompt("O>");
			break;
		default:
			cerr << "strange prompt code " << dat << endl;
		}
		stage = 0;
		// process queued commands
		if (! conn.more_line() && command_queue.size() > 0) {
			enter_command(command_queue.front());
			command_queue.pop_front();
		}
		break;
	case 2: // beep
		cons.set_prompt("BEEP");
		gdk_beep();
		stage = 0;
		break;
	case 7: // game info
		stage = 0;
		parser = & Igs::ps_game;
		cons.set_prompt("games");
		// drop the first line
		break;
	case 8: // file
		cons.set_prompt("file");
		stage = 8;
		parser = & Igs::ps_file;
		break;
	case 9: // ...
		if (rx.exec(dat, "Removing game (\\d+) from ob")) {
			data.observe_remove(atmi(0, 0));
		} else if (rx.exec(dat, "\\{(Game \\d+:.*)\\}")) {
			string d = atms(0, 0);
			sp_game_result(d);
		} else if (rx.exec(dat, "Adding game to")) {
		} else if (rx.exec(dat, "Set quiet to be")) {
		} else if (rx.exec(dat, "\\!\\!(.*)\\!\\!\\: (.*)")) {
			cons.dump(dat, IgsConsole::TT_SYSTEM);
		} else if (rx.exec(dat, "Game \\d+:")) {
			sp_game_result(dat);
		} else if (rx.exec(dat, "You can check .* finished.")) {
			data.score_match();
		} else if (rx.exec(dat, "Board is restored to .*")) {
			data.score_restore();
		} else if (rx.exec(dat, "Game has been adjourned.")) {
			data.game_adjourn();
		} else if (rx.exec(dat, "Removing @ (([A-H]|[J-T])\\d+)")) {
			data.score_remove(Place(atms(0, 0)));
		} else if (rx.exec(dat, "Player:\\s+\\S")) {
			parser = & Igs::ps_stats;
			stage = 0;
			ps_stats(s);
		} else if (rx.exec(dat, "Match\\[\\d+x\\d+\\] in \\d+ min")) {
			parser = & Igs::ps_match;
			stage = 0;
			ps_match(s);
		} else {
			// since its usually important...
			cons.dump(dat, IgsConsole::TT_SYSTEM);
		}
		break;
	case 11:
		stage = 0;
		parser = & Igs::ps_kibitz;
		ps_kibitz(s);
		break;
	case 15:
		stage = 0;
		parser = & Igs::ps_move;
		ps_move(s);
		break;
	case 18: // stored
		if (stage != 18) {
			stage = 18;
			// cons.dump("<Stored Games>", IgsConsole::TT_FILE);
		}
		cons.dump(dat, IgsConsole::TT_FILE);
		break;
	case 19: // say...
		if (rx.exec(dat, "\\*(\\S+)\\*: (.*)")) {
			// cons.dump(dat);
			data.add_talk(dat);
		} else {
			goto DUMP_STUFF;
			// cerr << "(19):" << dat << endl;
		}
		break;
	case 21: // misc
		if (rx.exec(dat, "\\{(Game \\d+:.*)\\}")) {
			string d = atms(0, 0);
			sp_game_result(d);
		} else if (rx.exec(dat, "\\{Match (\\d+): (\\S+) \\[.+\\]"
				   " vs. (\\S+)")) {
			data.game_update(atmi(0, 0), atms(0, 1), atms(0, 2));
		} else if (rx.exec(dat, "\\{Game (\\d+): (\\S+) vs (\\S+) "
				   "@ Move")) {
			data.game_update(atmi(0, 0), atms(0, 1), atms(0, 2));
		} else if (rx.exec(dat, "\\{(\\S+) \\[(.*)\\] has conn")) {
			data.player_update(atms(0, 0), atms(0, 1));
		} else if (rx.exec(dat, "\\{(\\S+) has discon")) {
			data.player_remove(atms(0, 0));
		} else if (rx.exec(dat, "\\!\\!(.*)\\!\\!\\: (.*)")) {
			// system message
			cons.dump(dat, IgsConsole::TT_SYSTEM);
		} else if (rx.exec(dat, "\\!(.*)\\!\\: (.*)")) {
			// yelling
			cons.dump(dat, IgsConsole::TT_YELL);
		} else {
			// cerr << "(21):" << dat << endl;
			goto DUMP_STUFF;
		}
		break;
	case 22:
		stage = 0;
		parser = & Igs::ps_board;
		ps_board(s);
		break;
	case 24: // tell...
		if (rx.exec(dat, "\\*(\\S+)\\*:  (.*)")) {
			cons.dump(dat);
		} else {
			// cerr << "(24):" << dat << endl;
			goto DUMP_STUFF;
		}
		break;
	case 25: // game result list;;;
		stage = 25;
		parser = & Igs::ps_file;
		break;
	case 26: // game time
		stage = 0;
		// parser = & Igs::ps_time;
		cons.dump(dat);
		break;
	case 27: // who
		stage = 0;
		parser = & Igs::ps_player;
		cons.set_prompt("who");
		// drop the first line
		break;
	case 28: // undo, since it's short
		if (rx.exec(dat, "Undo in game (\\d+):")) {
			data.undo_move(atmi(0, 0));
		} else if (rx.exec(dat, "\\S+ undid the")) {
			// in a match
			data.undo_move();
		} else {
			// cerr << "(28):" << dat << endl;
			goto DUMP_STUFF;
		}
		break;
	default:
	DUMP_STUFF:
		ostringstream t;
		t << "(" << cmd << ")=" << dat;
		cons.dump(t.str());
	}
}

void Igs::ps_game(const string & s)
{
	if (rx.exec(s, "7\\s+(.*)") == 0) {
		parser = & Igs::ps_prompt;
		ps_prompt(s);
		return;
	}
	if (stage == 0) {
		// new "game" info
		stage = 1;
	}
	string dat = atms(0, 0);
	try {
		GameData d(dat);
		data.player_update(d.white, d.w_rank, d.number);
		data.player_update(d.black, d.b_rank, d.number);
		data.game_update(d.number, d);
	} catch (string msg) {
		cerr << msg << endl;
	}
}

void Igs::ps_file(const string & s)
{
	if (rx.exec(s, "^(\\d+) ")) {
		parser = & Igs::ps_prompt;
		if (atmi(0, 0) != stage) ps_prompt(s);
		return;
	}
	if (stage == 25) {
		cons.dump("<game results>", IgsConsole::TT_FILE);
		stage = 0;
	}
	cons.dump(s, IgsConsole::TT_FILE);
}

void Igs::ps_move(const string & s)
{
	if (rx.exec(s, "15\\s+(.*)") == 0) {
		parser = & Igs::ps_prompt;
		if (rx.exec(s, "9 Game is titled: (.*)") == 0) ps_prompt(s);
		else data.set_title(current_game, atms(0, 0));
		return;
	}
	string d = atms(0, 0);
	// cerr << "Move Info: " << d << endl;
	if (stage == 0 &&
	    rx.exec(d, "^Game (\\d+) .: \\S+ \\(\\d+ (\\d+) (\\S+)\\)"
		    " vs \\S+ \\(\\d+ (\\d+) (\\S+)\\)") == 1) {
		// cerr << "game header" << endl;
		stage = 1;
		current_game = atmi(0, 0);
		data.set_time(current_game, atmi(0, 1), atmi(0, 2),
			      atmi(0, 3), atmi(0, 4));
	} else if (stage == 1) {
		if (rx.exec(d, "(\\d+)\\((\\S)\\)\\: (.+)") == 0) {
			return;
		}
		GoGame::Move m;
		int nm = atmi(0, 0);
		m.tn = atms(0, 1) == "B" ? GoGame::B_TURN : GoGame::W_TURN;
		string mv = atms(0, 2);
		// cerr << "processing move: " << mv;
		int n = rx.exec(mv, "(([A-H]|[J-T])\\d+)",
				regexx::Regexx::global);
		if (n > 0) {
			m.t = GoGame::MOVE_PUT;
			m.p = Place(atms(0, 0));
			for (int i = 1; i < n; ++ i) {
				m.k.push_back(Place(atms(i, 0)));
			}
		} else if (rx.exec(mv, "Handi\\S+ (\\d)")) {
			m.t = GoGame::MOVE_HANDICAP;
			int h = atmi(0, 0);
			// cerr << "handicap = " << h << endl;
			static const int hx[] = {3, 15, 3, 15, 3, 15, 9, 9};
			static const int hy[] = {3, 15, 15, 3, 9, 9, 3, 15};
			if (h > 4 && h % 2 == 1) {
				-- h;
				m.k.push_back(Place(9, 9));
			}
			for (int i = 0; i < h; i ++)
				m.k.push_back(Place(hx[i], hy[i]));
		} else if (rx.exec(mv, "Pass")) {
			m.t = GoGame::MOVE_PASS;
		} else {
			cerr << "what else could the move be? " << mv << endl;
			return;
		}
		data.add_move(current_game, nm, m);
	} else {
		// cerr << "unexpected 15 " << d << endl;
	}
}

void Igs::ps_board(const string & s)
{
	// cerr << "processing board" << endl;
	if (rx.exec(s, "22 (.*)") == 0) {
		parser = & Igs::ps_prompt;
		ps_prompt(s);
		if (game_board != NULL) delete [] game_board;
		game_board = NULL;
		return;
	}
	string d = atms(0, 0);
	if (stage < 2) {
		if (rx.exec(d, "(\\S+)(?!\\!) \\S+ \\d+ \\d+ \\S+") == 0)
			return;
		if (stage == 0) {
			game_white = atms(0, 0);
			stage = 1;
		} else {
			game_black = atms(0, 0);
			stage = 2;
		}
	} else {
		if (rx.exec(d, "(\\d+): (\\d+)") == 0) return;
		if (stage != 2 + atmi(0, 0)) return;
		string c = atms(0, 1);
		if (stage == 2) {
			if (game_board != NULL) delete [] game_board;
			game_size = c.length();
			game_board = new int[game_size * game_size];
		}
		if (int(c.length()) != game_size) return;
		for (int i = 0; i < int(c.length()); i ++) {
			game_board[i + (stage - 2) * game_size] = c[i] - '0';
		}
		stage ++;
		if (stage == game_size + 2) {
			// cerr << "board completed" << endl;
			parser = & Igs::ps_prompt;
			delete [] game_board;
			game_board = NULL;
		}
	}
}

void Igs::ps_player(const string & s)
{
	if (rx.exec(s, "27  (.*)") == 0) {
		parser = & Igs::ps_prompt;
		ps_prompt(s);
		return;
	}
	if (stage == 0) {
		// new "who" info
		stage = 1;
	}
	string d = atms(0, 0);
	int n = rx.exec(d, "(Q| |S)(!|X| )(\\d{3}| \\d{2}|  \\d| --)  "
			"(\\d{3}| \\d{2}|  \\d| --) (\\S+)\\s+(\\d+[ms])\\s+"
			"(\\S+)", regexx::Regexx::global);
	for (int i = 0; i < n; i ++) {
		char c = * atms(i, 1).c_str();
		PlayerData::PlayerStates s;
		switch (c) {
		case '!':
			s = PlayerData::PS_LOOK;
			break;
		case ' ':
			s = PlayerData::PS_OPEN;
			break;
		case 'X':
			s = PlayerData::PS_CLOSE;
			break;
		}
		data.player_update(atms(i, 4), atms(i, 6), s, atmi(i, 3),
				   atmi(i, 2), atms(i, 5));
	}
}

void Igs::ps_stats(const string & s)
{
	if (stage == 0) {
		stage = 1;
	}
	if (rx.exec(s, "9 (.*)") == 0) {
		parser = & Igs::ps_prompt;
		ps_prompt(s);
		return;
	}
	string d = atms(0, 0);
	if (rx.exec(d, "^Verbose  Bell  Quiet")) { // this is myself
		stage = 2;
	}
	if (stage == 1) cons.dump(d);
	else if (stage == 2) cons.dump(d, IgsConsole::TT_FIXED);
}

void Igs::ps_kibitz(const string & s)
{
	if (rx.exec(s, "11 (.*)") == 0) {
		parser = & Igs::ps_prompt;
		ps_prompt(s);
		return;
	}
	if (stage == 0) {
		string d = atms(0, 0);
		if (rx.exec(d, "Kibitz (\\S+ \\[.+\\]): Game \\S+ vs \\S+ \\["
			    "(\\d+)\\]") == 0) {
			cerr << "Kibitz head error" << endl;
			return;
		}
		stage = 1;
		talker = atms(0, 0);
		current_game = atmi(0, 1);
	} else if (stage == 1) {
		if (rx.exec(s, "11    (.+)") == 0) {
			cerr << "Kibitz word error" << endl;
		}
		// cerr << "talker is " << talker << endl;
		string w = talker + ": " + atms(0, 0);
		// cerr << "s=" << s;
		data.add_talk(current_game, w);
	}
}

void Igs::ps_match(const string & s)
{
	if (stage == 0) {
		stage = 1;
		return;
	}
	if (rx.exec(s, "9 Use <(.*)> or <(.*)> to respond") == 0) {
		parser = & Igs::ps_prompt;
		ps_prompt(s);
		return;
	}
	string acc = atms(0, 0);
	string dec = atms(0, 1);
	rx.exec(acc, "match (\\S+) (B|W) (\\d+) (\\d+) (\\d+)");
	string opp = atms(0, 0);
	GoGame::Turns t = (* atms(0, 1).c_str() == 'B' ? GoGame::B_TURN :
			   GoGame::W_TURN);
	IgsMatch * m = manage(new IgsMatch(opp, t, atmi(0, 2), atmi(0, 3),
					   atmi(0, 4)));
	m->accept_match.connect
		(SigC::bind<string>(slot(this, & Igs::float_command), acc));
	m->decline_match.connect
		(SigC::bind<string>(slot(this, & Igs::float_command), dec));
	m->show();
	//
	cons.dump(opp + " [" + data.find_rank(opp) + "] has");
	queue_command(string("stored -") + opp);
	queue_command(string("prob ") + opp);
}

void Igs::sp_game_result(const string & s)
{
	if (rx.exec(s, "Game (\\d+): \\S+ vs \\S+ (.*)") == 0) return;
	int n = atmi(0, 0);
	string r = atms(0, 1);
	if (rx.exec(r, "has adjourned")) data.game_adjourn(n);
	else if (rx.exec(r, ": (.*)")) data.game_finish(n, atms(0, 0));
	else if (rx.exec(r, "@")) return; // continued game...
	else {
		cerr << "strange game result: " << s << endl;
	}
}

void Igs::enter_command(const string & s)
{
	conn.send_line(s);
	if (catch_enter != NULL) {
		* catch_enter = s;
		catch_enter = NULL;
	}
}

void Igs::float_command(string s)
{
	enter_command(s);
}

void Igs::grab_line(const string & s)
{
	(this->* parser)(s);
}

void Igs::game_query(int n)
{
	ostringstream t;
	t << "games " << n;
	conn.send_line(t.str());
}

void Igs::moves_query(int n)
{
	ostringstream t;
	t << "moves " << n;
	conn.send_line(t.str());
}

void Igs::unob_game(int n)
{
	ostringstream t;
	t << "unob " << n;
	conn.send_line(t.str());
}

void Igs::observe_game(int n)
{
	ostringstream t;
	t << "observe " << n;
	conn.send_line(t.str());
}

void Igs::stats_player(const string & n)
{
	ostringstream t;
	t << "stats " << n;
	conn.send_line(t.str());
}

void Igs::send_move(const Place & p)
{
	conn.send_line(p);
}

void Igs::send_undo()
{
	conn.send_line("undo");
}

void Igs::send_pass()
{
	conn.send_line("pass");
}

void Igs::send_done()
{
	conn.send_line("done");
}

void Igs::send_say(const string & s)
{
	string t = "say ";
	t += s;
	conn.send_line(t);
}

void Igs::send_kibitz(const string & s, int n)
{
	stringstream t;
	t << "kibitz " << n << " " << s;
	conn.send_line(t.str());
}

void Igs::game_list_query()
{
	queue_command("games");
}

void Igs::player_list_query()
{
	queue_command("who");
}

void Igs::queue_command(const string & s)
{
	for (list<string>::iterator i = command_queue.begin();
	     i != command_queue.end(); i ++) if (s == * i) return;
	command_queue.push_back(s);
}
